
commands = commandArgs(T)

print("Loaded R File")
filePath <- commands[1] 
amountOfFiles <- strtoi(commands[2])
size <- strtoi(commands[3])

wrap.sequence <- function(sequence, lineLength){
  seq_length <- nchar(sequence)
  count <- ceiling(seq_length/lineLength)
  cutSequences <- array(dim = count)
  newSequence <- ""
  
  for(i in 1:count){
    cutSequences[i] <- substring(sequence, 1, lineLength)
    newSequence <- paste(newSequence, substring(sequence, 1, lineLength), "\n", sep="")
    sequence <- sub(substring(sequence, 1, lineLength), "", sequence)
  }
  return(newSequence)
}

LCSmedianValues <- vector(mode = "integer", length = amountOfFiles)
HOMmedianValues <- vector(mode = "integer", length = amountOfFiles)
LCSupperValues <- vector(mode = "integer", length = amountOfFiles)
HOMupperValues <- vector(mode = "integer", length = amountOfFiles)
LCSdataFromFile <- array(dim=c(amountOfFiles, (size*size-size)/2))
HOMdataFromFile <- array(dim=c(amountOfFiles, (size*size-size)/2))
SequenceSource <- scan(file = paste(filePath, "eval.txt", sep=""), 
                           what="char", sep="\n")


for(i in 1:amountOfFiles){
  lcsname <- paste(filePath, i, "_lcs.txt", sep="")
  homname <- paste(filePath, i, "_hom.txt", sep="")
  lcsdata = read.table(lcsname)
  homdata = read.table(homname)
  print(paste("Read file", i))
  
  pointer = 1
  LCSArray <- array()
  HOMArray <- array()
  for(g in 1:size){
    for(h in 1:g){
      if(g!=h){
        LCSArray[pointer] <- lcsdata[g,h]
        HOMArray[pointer] <- homdata[g,h]
        pointer = pointer +1
      }
    }
  }
  LCSdataFromFile[i,] <- LCSArray
  HOMdataFromFile[i,] <- HOMArray
  LCSmedianValues[i] <- boxplot(LCSArray, plot=F)$stats[3]
  LCSupperValues[i] <- boxplot(LCSArray, plot=F)$stats[4]
  HOMmedianValues[i] <- boxplot(HOMArray, plot=F)$stats[3]
  HOMupperValues[i] <- boxplot(HOMArray, plot=F)$stats[4]
}


outputPath <- strsplit(filePath, "_temp")
outputPath <- outputPath[[1]][1]
pdf(file = paste(outputPath, "/sequence_suggestion_lcs.pdf", sep=""))
chosenOne <- sample(which(LCSmedianValues == min(LCSmedianValues)), 1)
information <- SequenceSource[chosenOne*2-1]
information <- wrap.sequence(information, 74)
sequence <- SequenceSource[chosenOne*2-2]
sequence <- wrap.sequence(sequence, 60)

layout(matrix(c(3,3,1,2,1,2), 3, 2, byrow = TRUE))
par(mar=c(5, 5, 2, 2))
boxplot(LCSdataFromFile[chosenOne,], LCSdataFromFile[-chosenOne,], 
        names=c("suggested sequence", "all sequences"), ylab="bp", 
        cex.lab=1.5, cex.axis=1.3, col=(c("green2","grey88")))
title("longest common substring")
boxplot(HOMdataFromFile[chosenOne,], HOMdataFromFile[-chosenOne,], 
        names=c("suggested sequence", "all sequences"), ylab="bp", 
        cex.lab=1.5, cex.axis=1.3, col=(c("green2","grey88")))
title("homology")

par(mar=c(0, 2, 2, 0))
plot.default(c(0,20), c(0,20), type="n", axes=F,
             ylab="", xlab="")
text(-0.5, 20, "Suggested sequence - lowest homology", cex=3, adj=c(0,1))
text(-0.5, 15, information, cex=1.2, adj=c(0,1), family="mono")
text(-0.5, 11, sequence, cex=1.5, adj=c(0,1), family="mono")
dev.off()


pdf(file = paste(outputPath, "/sequence_suggestion_hom.pdf", sep=""))
chosenOne <- sample(which(HOMmedianValues == min(HOMmedianValues)), 1)
information <- SequenceSource[chosenOne*2-1]
information <- wrap.sequence(information, 74)
sequence <- SequenceSource[chosenOne*2-2]
sequence <- wrap.sequence(sequence, 60)

layout(matrix(c(3,3,1,2,1,2), 3, 2, byrow = TRUE))
par(mar=c(5, 5, 2, 2))
boxplot(LCSdataFromFile[chosenOne,], LCSdataFromFile[-chosenOne,], 
        names=c("suggested sequence", "all sequences"), ylab="bp", 
        cex.lab=1.5, cex.axis=1.3, col=(c("green2","grey88")))
title("longest common substring")
boxplot(HOMdataFromFile[chosenOne,], HOMdataFromFile[-chosenOne,], 
        names=c("suggested sequence", "all sequences"), ylab="bp", 
        cex.lab=1.5, cex.axis=1.3, col=(c("green2","grey88")))
title("homology")

par(mar=c(0, 2, 2, 0))
plot.default(c(0,20), c(0,20), type="n", axes=F,
             ylab="", xlab="")
text(-0.5, 20, "Suggested sequence - lowest homology", cex=3, adj=c(0,1))
text(-0.5, 15, information, cex=1.2, adj=c(0,1), family="mono")
text(-0.5, 11, sequence, cex=1.5, adj=c(0,1), family="mono")
dev.off()